#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n",
        "--lineno",
        dest="lineno",
        default=False,
        action="store_true",
        help="show line no.",
    )
    parser.add_argument("-f", "--file", dest="tgtfile", help="file to check")
    parser.add_argument(
        "-r",
        "--rows",
        dest="rows",
        default=999,
        type=int,
        help="how many rows with nonascii char to be dumped",
    )
    parser.add_argument(
        "-a",
        "--all",
        dest="all",
        action="store_true",
        help="show lines without nonascii characters",
    )
    parser.add_argument(
        "-p",
        "--pipe",
        dest="pipe",
        action="store_true",
        default=False,
        help="input from pipe",
    )
    parser.add_argument(
        "-P",
        "--pipefile",
        dest="pipefile",
        action="store_true",
        default=False,
        help="treat each line from pipe as potential file to scan.",
    )
    parser.add_argument(
        "-L",
        "--followlink",
        dest="link",
        action="store_true",
        default=False,
        help="follow link",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    if not (args.tgtfile or args.pipe or args.pipefile):
        parser.print_help()
        sys.exit(0)

    def scan_stream(INPUT, pipefile=False):
        if not INPUT:
            print("")
            return
        n = 0
        found = 0
        for ln in INPUT:
            if found >= args.rows:
                break
            ln = ln.rstrip()
            n += 1
            nonasciis = [
                (ix, c, str(hex(ord(c)))) for ix, c in enumerate(ln) if ord(c) > 127
            ]
            if nonasciis or args.all:
                found += 1
                if args.tgtfile and found == 1:
                    print(" -> contains nonascii")
                if args.lineno:
                    if nonasciis:
                        print("  {}|{} => {}".format(n, ln, nonasciis))
                    else:
                        print("  {}|{}".format(n, ln))
                else:
                    if nonasciis:
                        print("  {} => {}".format(ln, nonasciis))
                    else:
                        print("  {}".format(ln))
            if pipefile and os.path.isfile(ln):
                with open(ln, "r") as fin:
                    print("# scanning {}".format(ln))
                    scan_stream(fin.readlines(), pipefile=False)

        if found == 0:
            print("")

    filename = None
    INPUT = None
    if args.pipe or args.pipefile:
        INPUT = sys.stdin
    if args.tgtfile:
        nonasciis = [
            (ix, c, str(hex(ord(c))))
            for ix, c in enumerate(args.tgtfile)
            if ord(c) > 127
        ]
        if nonasciis:
            print("  {} ({})".format(args.tgtfile, nonasciis), end="")
        else:
            print("  {}".format(args.tgtfile), end="")
        if os.path.isfile(args.tgtfile) and (
            not os.path.islink(args.tgtfile) or args.link
        ):
            with open(args.tgtfile, "r") as fin:
                INPUT = fin.readlines()
                filename = args.tgtfile
    scan_stream(INPUT, pipefile=args.pipefile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang

import os
import sys
import argparse
import re
import json
import readline
import string
from datetime import datetime
from time import sleep
from texttable import Texttable
from sretools import SimpleTable

args = None
db = dict()


def initdb():
    global args
    global db
    args.db = os.path.expanduser(args.db)
    if not os.path.isdir(os.path.dirname(args.db)):
        os.mkdir(os.path.dirname(args.db))
    if not os.path.isfile(args.db):
        with open(args.db, "w") as f:
            f.write(json.dumps(db, indent=2, sort_keys=True))
    else:
        with open(args.db, "r") as f:
            db = json.loads(f.read())
    return db


def closedb():
    global args
    global db
    with open(args.db, "w") as f:
        f.write(json.dumps(db, indent=2, sort_keys=True))


def show_incident_list():
    global args
    global db
    rows = list()
    for inum, inc in db.items():
        rows.append(
            [
                str(inc.get("incident_number")),
                inc.get("create_time"),
                inc.get("brief"),
                str(len(inc.get("updates"))),
            ]
        )
    table = SimpleTable(
        header=["incident#", "create_time", "brief", "#updt"], data=rows
    )
    print(table)


def delete_incident():
    global args
    global db
    if args.incident not in db:
        print("# cannot find {}".format(args.incident))
        sys.exit(-1)
    del db[args.incident]
    closedb()


def show_incident():
    global args
    global db
    if args.incident not in db:
        print("# cannot find {}".format(args.incident))
        sys.exit(-1)
    # print(json.dumps(db[args.incident],indent=2,sort_keys=True))
    cout = ""
    mout = ""
    wout = "<html>\n<header>\n"
    wout += "<tt>\n"
    i = db[args.incident]
    mout += "## {}\n".format(i["brief"])
    mout += "### {} \n".format(i["incident_number"])
    mout += "| __meta__ | __value__ |\n"
    mout += "| ---- | ---- |\n"
    cout += "# ({}) {}\n\n".format(i["incident_number"], i["brief"])
    wout += "<title>{} Postmortem</title>\n".format(i["incident_number"])
    wout += "<h2>{} - {}</h2>\n".format(i["incident_number"], i["brief"])
    wout += "</header>\n"
    wout += "<table>\n"
    mout += "|{}|{}|\n".format("CreateTime", i["create_time"])
    cout += "Create_time :\n    {}\n".format(i["create_time"])
    wout += "<tr><td><b>CreateTime</b></td><td>{}</td></tr>\n".format(i["create_time"])
    for keym in [
        "OncallOps",
        "Status",
        "Impact",
        "RootCause",
        "DurationOfProblem",
        "ServiceImpacted",
        "%Impacted",
        "UserImpact",
        "RevenueImpact",
        "HowToRepeat",
        "Resolution",
        "References",
        "LessonLearned",
        "ActionItem",
        "KeywordsForSearch",
    ]:
        if keym in i and i[keym]:
            cout += "{} : \n".format(keym)
            cout += "    {}\n".format(i[keym])
            wout += "<tr><td><b>{}</b></td><td>{}</td></tr>\n".format(keym, i[keym])
            mout += "| {} | {} |\n".format(keym, i[keym])
        else:
            if args.post:
                i[keym] = input("# {}=".format(keym))
                cout += "{} : \n".format(keym)
                cout += "    {}\n".format(i[keym])
                wout += "<tr><td><b>{}</b></td><td>{}</td></tr>\n".format(keym, i[keym])
                mout += "| {} | {} |\n".format(keym, i[keym])
                closedb()
    wout += "</table>\n"
    if args.json:
        print(json.dumps(i, indent=2, sort_keys=True))
        return
    cout += "\n\n# timeline   : \n"
    wout += "<h2>Timeline    :</h2>\n"
    mout += "\n\n"
    mout += "### {} \n".format("Timeline :")
    mout += "| __datetime__ | __brief__ | __detail__ | \n"
    mout += "| ---- | ---- | ---- | \n"
    # wout += "<table style=\"border-collapse: collapse;border: 1px;\">\n"
    wout += '<table border=1 style="border-collapse: collapse;">\n'
    wout += '<tr style="background-color:#F3F3F3"><th>update_time</th><th>brief</th><th>detail</th></tr>\n'
    table = Texttable()
    table.set_header_align(["l", "l", "l"])
    table.set_cols_align(["l", "l", "l"])
    table.set_chars([" ", " ", " ", "-"])
    maxw = 0
    for x in [u.get("xcontent", "") for u in i.get("updates", dict())]:
        for l in x.splitlines():
            if len(l) > maxw:
                maxw = len(l)
    table.set_cols_width([12, 25, maxw + 1])
    rows = list()
    rows.append(["update_time", "brief", "detail"])
    for u in sorted(i.get("updates", []), key=lambda x: x["create_time"], reverse=True):
        rows.append([u["create_time"][:-3], u["update_content"], u.get("xcontent", "")])
        mout += "| {} | {} | <pre>{}</pre> | \n".format(
            u["create_time"][:-3],
            u["update_content"],
            re.sub(
                r"([\\\`\*\_\{\}\[\]\(\)#+-.!|])",
                r"\\\1",
                u.get("xcontent", "").replace("\n", "</br>").replace("~", "\$HOME"),
            ),
        )
        wout += "<tr>\n"
        for val in [u["create_time"][:-3], u["update_content"], u.get("xcontent", "")]:
            wout += "<td>{}</td>\n".format(
                val.replace("\n", "</br>").replace(" ", "&nbsp;")
            )
        wout += "</tr>\n"

    mout += "\n\n"
    table.add_rows(rows)
    stbl = table.draw()
    stbl = "\n".join([re.sub(r"\s+$", "", ln) for ln in stbl.splitlines()])
    cout += stbl + "\n"
    wout += "</table>\n"
    wout += "</html>\n"
    if args.export:
        if args.expand:
            incdir = (
                args.incident
                + "."
                + "".join(
                    c
                    for c in re.sub(r"\s+", ".", i["brief"])
                    if c in set(string.ascii_letters + string.digits + "_-.")
                )
            )
        else:
            incdir = args.incident
        if not os.path.exists(incdir):
            os.mkdir(incdir)
        with open("{}/{}.html".format(incdir, args.incident), "w") as f:
            f.write(wout)
        with open("{}/{}.txt".format(incdir, args.incident), "w") as f:
            f.write(cout)
        with open("{}/{}.md".format(incdir, args.incident), "w") as f:
            f.write(mout)
        with open("{}/{}.json".format(incdir, args.incident), "w") as f:
            f.write(json.dumps(db[args.incident], indent=2, sort_keys=True))
        sys.exit(0)
    if args.html:
        print(wout)
    elif args.markdown:
        print(mout)
    else:
        print(cout)


def add_new_incident():
    global args
    global db
    now = datetime.now()
    snow = "c" + now.strftime("%Y%m%d%H%M%S")
    create_time = now.strftime("%Y-%m-%d %H:%M:%S")
    while snow in db:
        sleep(1)
        now = datetime.now()
        snow = "c" + now.strftime("%Y%m%d%H%M%S")
        create_time = now.strftime("%Y-%m-%d %H:%M:%S")
    incident = dict()
    incident["incident_number"] = snow
    incident["create_time"] = create_time
    incident["brief"] = args.new
    incident["updates"] = list()
    db[snow] = incident
    print("# incident {} added.".format(snow))
    closedb()
    return snow


def update_incident():
    global args
    global db
    if args.incident not in db:
        print("# cannot find {}".format(args.incident))
        sys.exit(-1)
    updt = dict()
    now = datetime.now()
    update_time = now.strftime("%Y-%m-%d %H:%M:%S")
    updt["create_time"] = update_time
    updt["update_content"] = args.update
    updt["xcontent"] = args.attach
    if args.attach == "vi":
        tmpfile = "/tmp/.incident.{}.{}".format(
            args.incident, now.strftime("%Y%m%d%H%M%S")
        )
        os.system("vi {}".format(tmpfile))
        with open(tmpfile, "r") as f:
            updt["xcontent"] = f.read()
        os.unlink(tmpfile)
    if os.path.isfile(args.attach):
        with open(args.attach, "r") as f:
            updt["xcontent"] = f.read()
    db[args.incident]["updates"].append(updt)
    closedb()


def main():
    global args
    global db
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--dbfile",
        dest="db",
        default="~/.sretools/incidents.json",
        action="store_true",
        help="db file",
    )
    parser.add_argument(
        "-R",
        "--report",
        dest="report",
        default=False,
        action="store_true",
        help="run report",
    )
    parser.add_argument("-i", "--incident", dest="incident", help="incident number")
    parser.add_argument(
        "-n", "--new_incident", dest="new", help="new incident. value used for brief"
    )
    parser.add_argument(
        "-u", "--update_incident", dest="update", help="update incident with progress"
    )
    parser.add_argument(
        "-x",
        "--delete_incident",
        dest="delete",
        action="store_true",
        default=False,
        help="delete incident",
    )
    parser.add_argument(
        "-a",
        "--attachment",
        dest="attach",
        default="",
        help="detail info from file or vi.",
    )
    parser.add_argument(
        "-J",
        "--json",
        dest="json",
        action="store_true",
        default=False,
        help="dump incident postmortem in JSON format",
    )
    parser.add_argument(
        "-p",
        "--postmortem",
        dest="post",
        action="store_true",
        default=False,
        help="google SRE postmortem format",
    )
    parser.add_argument(
        "-w",
        "--html",
        dest="html",
        action="store_true",
        default=False,
        help="dump report in HTML format",
    )
    parser.add_argument(
        "-m",
        "--markdown",
        dest="markdown",
        default=False,
        action="store_true",
        help="dump report in Markdown format",
    )
    parser.add_argument(
        "-F",
        "--export",
        dest="export",
        action="store_true",
        default=False,
        help="export json/html/markdown file as well. use with -p.",
    )
    parser.add_argument(
        "-e",
        "--noexpanddirname",
        dest="expand",
        action="store_false",
        default=True,
        help="with -F. only use cid for dirname when specified.",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()
    if args.new:
        args.new = re.sub(r"\\n", "\n", args.new)
    if args.update:
        args.update = re.sub(r"\\n", "\n", args.update)
    if args.attach:
        args.attach = re.sub(r"\\n", "\n", args.attach)
    if not (args.incident or args.new or args.report):
        parser.print_help()
        sys.exit(0)
    db = initdb()
    if args.report:
        if not args.incident:
            show_incident_list()
        else:
            show_incident()
        sys.exit(0)
    if args.delete:
        delete_incident()
        closedb()
    if args.new:
        add_new_incident()
        closedb()
    if args.incident and args.update:
        update_incident()
        closedb()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import re
import json
import yaml
import random
import string


class JsonConverter:
    def __init__(
        self,
        jsstr,
        tblattr="border-collapse:collapse;",
        sort=False,
        sort_by_val=False,
        sortkeywords=None,
        recursive=False,
        keys_included=None,
        keys_excluded=None,
        colors=None,
        collapseandexpand=False
    ):
        if type(jsstr) is list or type(jsstr) is dict:
            self.__obj = jsstr
        else:
            try:
                self.__obj = json.loads(jsstr)
            except:
                print("# invalid JSON.")
                sys.exit(0)
        if type(self.__obj) is dict:
            if keys_included:
                self.__obj = {k: v for k, v in self.__obj.items() if k in keys_included}
            if keys_excluded:
                for keys in keys_excluded:
                    klst = re.split(r"\.", keys)
                    pt = self.__obj
                    good = True
                    for k in klst[:-1]:
                        if k not in pt:
                            print(
                                "# warning. ignored invalid key series: {}".format(keys)
                            )
                            good = False
                            break
                        pt = pt[k]
                    if good:
                        del pt[klst[-1]]
        self.__recursive = recursive
        self.__sort = sort
        self.__sort_by_val = sort_by_val
        self.__tblattr = tblattr
        self.__sortkeywords = sortkeywords
        if colors:
            self.__colors = re.split(",", colors)
        else:
            self.__colors = None
        self.__lastcolor = 0
        self.__collapseandexpand = collapseandexpand

    def json2html(self):
        return self.json2html_helper(self.__obj)

    def __is_final_list(self, lst):
        if not type(lst) is list:
            return False
        for r in lst:
            if type(r) is dict or type(r) is list:
                return False
        return True

    def __is_final_tbl(self, lst):
        if not type(lst) is list:
            return False
        for r in lst:
            if not self.__is_final_dt(r):
                return False
        return True

    def __is_final_dt(self, dt):
        if not type(dt) is dict:
            return False
        for k, v in dt.items():
            if type(v) is dict or type(v) is list:
                return False
        return True

    def json2yaml(self):
        return yaml.safe_dump(self.__obj, default_flow_style=False)

    @staticmethod
    def jsonize(p,depth=9999,keyset=None,explainlist=True) :
        if depth <= 0 :
            return p
        if type(p) is dict :
            for k,c in p.items() :
                if type(c) is str :
                    if (keyset and k in keyset) or (not keyset) :
                        try :
                            p[k] = JsonConverter.jsonize(json.loads(c),depth=depth-1,keyset=keyset,explainlist=explainlist)
                        except :
                            pass
                else :
                        p[k] = JsonConverter.jsonize(p[k],depth=depth-1,keyset=keyset,explainlist=explainlist)
        elif type(p) is list :
            for i,k in enumerate(p) :
                if type(k) is str :
                    if explainlist :
                        try :
                            p[i] = JsonConverter.jsonize(json.loads(k),depth=depth-1,keyset=keyset,explainlist=explainlist)
                        except :
                            pass
                else :
                        p[i] = JsonConverter.jsonize(p[i],depth=depth-1,keyset=keyset,explainlist=explainlist)
        elif type(p) is set :
            p = JsonConverter.jsonize(list(p))
        elif type(p) is str :
            try :
                p = JsonConverter.jsonize(json.loads(p),depth=depth-1,keyset=keyset,explainlist=explainlist)
            except :
                pass
        else :
            print("# unkown data type.")
            pass
        return p

    def json2html_helper(self, obj, lvl=0):
        out = ""
        inlist_border = 1
        ctoggle = '+/-'
        jqyhdr = """
<header>
<script src="https://code.jquery.com/jquery-3.6.0.min.js"></script> 
<script type="text/javascript"> 
    $(document).ready(function () { 
        $('tr.parent') .css("cursor", "pointer") .attr("title", "Click to expand/collapse") .click(function () { 
            $(this).siblings('.child-' + this.id).toggle(); 
        }); 
        $('tr[@class^=child-]').hide().children('td'); 
    }); 
</script>
</header>
"""
        curshift = " " * 4 * lvl
        if lvl == 0 :
            if self.__collapseandexpand :
                out += jqyhdr
            out += "\n<tt>\n" 

        if type(obj) is not list and type(obj) is not dict:
            if self.__recursive:
                try:
                    o = json.loads(obj)
                    out += self.json2html_helper(o,lvl+1)
                except:
                    out += curshift + str(obj)
            else:
                out += curshift + str(obj)
            return out
        if self.__colors:
            c = self.__colors[self.__lastcolor]
            self.__lastcolor = (self.__lastcolor + 1) % len(self.__colors)
            colstr = "background-color:{}".format(c)
        else:
            colstr = ""
        if type(obj) is list:
            if self.__sort:
                obj = sorted(obj)
            style = 'style="{};{}"'.format(self.__tblattr, colstr)
            if self.__is_final_list(obj):
                for r in obj:
                    r = str(r).replace("\r\n", "<br>")
                    r = str(r).replace("\n", "<br>")
                    out += curshift + "<li>" + str(r) + "</li>\n"
            elif self.__is_final_tbl(obj):
                tid = "".join([random.choice(string.ascii_letters) for _ in range(20)])
                tw = 'width="100%"' if lvl != 0 else ""
                out += curshift + "<table {} border={} {}>\n".format(style, inlist_border,tw)
                if self.__collapseandexpand and lvl != 0 :
                    #out += curshift + """<tr class="parent" id="{}" title="Click to expand/collapse" style="cursor: pointer;"> <td bgcolor="#FFFFCC">{}(L)</td> </tr>\n""".format(tid,ctoggle)
                    out += curshift + """<tr class="parent" id="{}" title="Click to expand/collapse" style="cursor: pointer;"> <td bgcolor="#FFFFCC">L({})</td> </tr>\n""".format(tid,len(obj))
                need_switch = len(obj) > 1
                colors = ["#FFFFFF", "F8F8F8"]
                ix = 0
                for r in obj:
                    c = colors[ix]
                    ix = 1 - ix
                    for k, v in r.items():
                        if not v:
                            v = ""
                        else:
                            v = str(v).replace("\r\n", "<br>")
                            v = str(v).replace("\n", "<br>")
                        if need_switch:
                            out += curshift + '<tr class="child-{}" style="background-color:{}">\n'.format(tid,c)
                        else:
                            out += curshift + "<tr>\n"
                        out += curshift + "    " + "<td valign=\"top\"><b>" + str(k) + "</b></td>\n"
                        out += curshift + "    " + "<td>" + str(v) + "</td>\n"
                        out += curshift + "</tr>\n"
                    # + sepbar
                    #out += curshift + "<tr><td></td><td></td></tr>\n"
                out += curshift + "</table>\n"
            else:
                tid = "".join([random.choice(string.ascii_letters) for _ in range(20)])
                tw = 'width="100%"' if lvl != 0 else ""
                out += curshift + "<table {} border={} {}>\n".format(style, inlist_border,tw)
                if self.__collapseandexpand and lvl != 0 :
                    #out += curshift + """<tr class="parent" id="{}" title="Click to expand/collapse" style="cursor: pointer;"> <td bgcolor="#FFFFCC">{}(L)</td> </tr>\n""".format(tid,ctoggle)
                    out += curshift + """<tr class="parent" id="{}" title="Click to expand/collapse" style="cursor: pointer;"> <td bgcolor="#FFFFCC">L({})</td> </tr>\n""".format(tid,len(obj))
                for o in obj:
                    out += curshift + '<tr class="child-{}">\n'.format(tid)
                    out += curshift + "    " + "<td valign=\"top\">\n"
                    out += self.json2html_helper(o,lvl+2)
                    out += "\n"
                    out += curshift + "    " + "</td>\n"
                    out += curshift + "</tr>\n"
                out += curshift + "</table>\n"
            return out
        if type(obj) is dict:
            if self.__sort:
                if self.__sort_by_val:
                    fs = lambda t: t[1]
                else:
                    fs = lambda t: t[0]
            else:
                fs = lambda t: t
            if self.__sortkeywords:
                def f(s):
                    stbl = {
                        s: ix
                        for ix, s in enumerate(re.split(r",", self.__sortkeywords))
                    }
                    return (stbl.get(s, 999999), s)

                fs = lambda t: f(t[0])
            style = 'style="{};{}"'.format(self.__tblattr, colstr)
            tid = "".join([random.choice(string.ascii_letters) for _ in range(20)])
            tw = 'width="100%"' if lvl != 0 else ""
            out += curshift + "<table {} border={} {}>\n".format(style, inlist_border,tw)
            if self.__collapseandexpand and lvl != 0 :
                #out += curshift + """<tr class="parent" id="{}" title="Click to expand/collapse" style="cursor: pointer;"> <td bgcolor="#FFFFCC">{}(D)</td> </tr>\n""".format(tid,ctoggle)
                out += curshift + """<tr class="parent" id="{}" title="Click to expand/collapse" style="cursor: pointer;"> <td bgcolor="#FFFFCC">D({})</td> </tr>\n""".format(tid,len(obj))
            for k, v in sorted(obj.items(), key=fs):
                out += curshift + '<tr class="child-{}">\n'.format(tid)
                out += curshift + "<td valign=\"top\"><b>" + str(k) + "</b></td>\n"
                out += curshift + "    " + "<td>\n"
                out += self.json2html_helper(v,lvl+2)
                out += "\n"
                out += curshift + "    " + "</td>\n"
                out += curshift + "</tr>\n"
            out += curshift + "</table>\n"
            return out


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--infile", dest="infile", help="input file")
    parser.add_argument(
        "-a",
        "--table_attributes",
        dest="attributes",
        default="border-collapse:collapse;",
        help="table attributes",
    )
    parser.add_argument(
        "-s",
        "--sorted",
        dest="sorted",
        action="store_true",
        default=False,
        help="try best to make sure output is sorted. default by key.",
    )
    parser.add_argument(
        "-C", "--colors", dest="colors", help="background colors to be used."
    )
    parser.add_argument(
        "-k",
        "--keywords_order",
        dest="keyorder",
        help="use this ordered keywords to help sort dict.",
    )
    parser.add_argument(
        "-I",
        "--keys-included",
        dest="keys_included",
        action="append",
        help="include these TOP level keys only",
    )
    parser.add_argument(
        "-E",
        "--keys-excluded",
        dest="keys_excluded",
        action="append",
        help="exclude these keys. eg. k1 or k2.k3",
    )
    parser.add_argument(
        "-v",
        "--dictsortbyvalue",
        dest="dtbyval",
        action="store_true",
        default=False,
        help="with -s, sort dict by value.",
    )
    parser.add_argument(
        "-Y",
        "--yaml",
        dest="yaml",
        action="store_true",
        default=False,
        help="dump in YAML format",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        dest="recursive",
        action="store_true",
        default=False,
        help="if str value is JSON, treat it as JSON object ",
    )
    parser.add_argument(
        "-Z",
        "--collapseandexpand",
        dest="collapseandexpand",
        action="store_true",
        default=False,
        help="make the HTML table able to expand/collapse. this may make it looks less formal.",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    if args.infile:
        if not os.path.isfile(args.infile):
            print("{} not exists.")
        with open(args.infile, "r") as f:
            INPUT = f.read()
    else:
        INPUT = sys.stdin.read()

    if args.yaml:
        print(
            JsonConverter(
                INPUT,
                tblattr=args.attributes,
                sort=args.sorted,
                sort_by_val=args.dtbyval,
                keys_included=args.keys_included,
                keys_excluded=args.keys_excluded,
                sortkeywords=args.keyorder,
            ).json2yaml()
        )
        sys.exit(0)

    print(
        JsonConverter(
            INPUT,
            tblattr=args.attributes,
            sort=args.sorted,
            sort_by_val=args.dtbyval,
            sortkeywords=args.keyorder,
            recursive=args.recursive,
            keys_included=args.keys_included,
            keys_excluded=args.keys_excluded,
            colors=args.colors,
            collapseandexpand=args.collapseandexpand,
        ).json2html()
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang

import os
import sys
import argparse
import re
import getpass
import pexpect
import string


class LE:
    def __init__(self, command, usedefaults=True, timeout=10, consts=None, rules=None):
        self.command = command or "echo"
        self.usedefaults = usedefaults
        self.timeout = timeout
        self.consts = consts or dict()
        self.rules = rules or list()
        self.debug = False
        if self.usedefaults:
            self.rules.append(("[\$#]", "%IGNORE"))
            self.rules.append(("%WHENEOF", "%BREAK"))
            self.rules.append(("%WHENTIMEOUT", "%NONE"))

    def add_const(self, const=None, value=None):
        if not (const and value):
            return
        self.consts[str(const)] = str(value)

    def add_rule(self, pattern, response):
        if not (pattern and response):
            return
        self.rules.append((str(pattern), str(response)))

    def set_debug(self, state=True):
        self.debug = state

    def spawn(self):
        for k, v in self.consts.items():
            if v == "INPUT":
                v = input("% " + k + " : ")
                self.consts[k] = v
            elif v == "PASSWORD":
                v = getpass.getpass("% " + k + " : ")
                self.consts[k] = v
        patterns = list()
        responses = list()
        for p, r in self.rules:
            if p == "%WHENEOF":
                p = pexpect.EOF
            elif p == "%WHENTIMEOUT":
                p = pexpect.TIMEOUT
            if r == "%NONE":
                r = None
            if r:
                m = re.search(r"\{\{(.*?)\}\}", r)
                while m and m.group(1) in self.consts:
                    k = m.group(1)
                    r = re.sub("\{\{" + k + "\}\}", self.consts[k], r)
                    m = re.search(r"\{\{(.*?)\}\}", r)
            patterns.append(p)
            responses.append(r)
        self.child = pexpect.spawn(self.command, timeout=self.timeout)
        cnt = 0
        rcnt = -999
        respawn = False
        while True:
            r = self.child.expect(patterns)
            cnt += 1
            if respawn and rcnt + 1 != cnt:
                respawn = False
            if self.debug:
                print("# Matched : {} (r={})".format(patterns[r], r))
            # timeout
            if patterns[r] == pexpect.TIMEOUT:
                self.print_content()
                print()
                break
            # EOF
            if patterns[r] == pexpect.EOF:
                self.print_content()
                print()
                if respawn:
                    if rcnt + 1 == cnt:
                        self.child.close()
                        self.child = pexpect.spawn(self.command, timeout=self.timeout)
                        continue
                print()
                break
            rsp = responses[r]
            if not rsp:
                break
            actlst = re.split(r";", rsp)
            if self.debug:
                print("# actlst = {}".format(actlst))
            self.print_content()
            gobreak = False
            for rsp in actlst:
                if rsp.upper() == "%IGNORE":
                    continue
                if rsp.upper() == "%BREAK":
                    gobreak = True
                    break
                if re.search(r"%CMD=(\S+)", rsp.upper()):
                    command = re.sub(r"%(CMD|cmd)=", "", rsp, count=1)
                    if self.debug:
                        print("# runcmd : {}".format(command))
                    os.system(command)
                if rsp.upper() == "%RESPAWN":
                    if self.debug:
                        print("# set respawn flag")
                    respawn = True
                    rcnt = cnt
                    continue
                if self.debug:
                    print("# sendline : {}".format(rsp))
                self.child.sendline(rsp)
            if gobreak:
                break
        self.child.interact()

    def print_content(self):
        if not self.child:
            return
        content = self.child.before.decode()
        if self.child.after:
            try:
                content += self.child.after.decode()
            except:
                pass
        printable = set(string.printable)
        if re.search(r"\S+", content):
            lines = content.splitlines()
            output = "\n".join(
                ("".join((x for x in ln if x in printable)) for ln in lines)
            )
            print(output, end="")

    def __repr__(self):
        output = "LE {\n"
        output += "  command : {}\n".format(self.command)
        for k, v in self.consts.items():
            output += "  const  {} : {}\n".format(str(k), str(v))
        for i in self.rules:
            output += "  rule    : {}\n".format(str(i))
        output += "}\n"
        return output


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--command", dest="command", help="command to spawn")
    parser.add_argument("--const", dest="consts", action="append", help="const->value")
    parser.add_argument(
        "--rule", dest="rules", action="append", help="pattern->response"
    )
    parser.add_argument(
        "--timeout", dest="timeout", type=int, default=60, help="timeout"
    )
    parser.add_argument("--debug", dest="debug", action="store_true", help="debug mode")
    args = parser.parse_args()
    if args.debug:
        print("# command = ", args.command)
        print("# consts  = ", args.consts)
        print("# rules   = ", args.rules)
    if not args.command:
        print("# no command specified.")
        parser.print_help()
        sys.exit(0)
    le = LE(command=args.command, timeout=args.timeout)
    le.set_debug(args.debug)
    for c in args.consts or []:
        m = re.search(r"(.+?)->(.+)", c)
        if m:
            le.add_const(m.group(1), m.group(2))
    for r in args.rules or []:
        m = re.search(r"(.+?)->(.+)", r)
        if m:
            le.add_rule(m.group(1), m.group(2))
    if args.debug:
        print("# LE = ", str(le))
    le.spawn()


if __name__ == "__main__":
    main()



import sys
import os
import subprocess

class commandline :
    @staticmethod
    def qx(cmd, merge=False, debug=False, exitonerror=False, hidepwd=False, pwdmark="") :
        if type(cmd) is not list :
            if type(cmd) is str :
                cmd = cmd.split()
            else :
                raise Exception("cmd must be a list or a string.")
        cmdmasked = list()
        if hidepwd :
            i=0
            while i < len(cmd) :
                if cmd[i] in ['-'+c for c in pwdmark] :
                    cmdmasked.append(cmd[i])
                    if i+1 < len(cmd) :
                        cmdmasked.append("******")
                    i+=2
                else :
                    cmdmasked.append(cmd[i])
                    i+=1
        if debug :
            print("# cmd=[{}]".format(" ".join(cmdmasked or cmd)))
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.subprocess.STDOUT if merge else subprocess.PIPE) as p :
            out, err = p.communicate()
            out = out and out.decode().rstrip()
            err = err and err.decode().rstrip()
            if debug :
                print("# stdout=[{}]".format(out or ""))
                print("# stderr=[{}]".format(err or ""))
                print("# rtcode=[{}]".format(p.returncode))
            if p.returncode != 0 and exitonerror :
                sys.exit(p.returncode)
            return (p.returncode, out, err)
            
        
#print(commandline.qx("echo hello -p xx yy",debug=True,hidepwd=True,pwdmark="p"))
#print(commandline.qx("ls -p",debug=True,hidepwd=True,pwdmark="p"))



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import re
import jaydebeapi
from sretools import SimpleTable


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-J", "--jar", dest="jar", help="jdbc driver jar file")
    parser.add_argument("-D", "--driver", dest="driver", help="jdbc driver name")
    parser.add_argument("-U", "--url", dest="url", help="jdbc url or connection string")
    parser.add_argument("-Q", "--sql", dest="sql", help="sql statement to run")
    parser.add_argument(
        "-w",
        "--maxwidth",
        dest="maxwidth",
        type=int,
        default=-1,
        help="maxwidth of column",
    )
    parser.add_argument(
        "-F", "--outformat", dest="format", default="", help="json,yaml,csv"
    )
    parser.add_argument(
        "-v",
        "--pivot",
        dest="pivot",
        action="store_true",
        default=False,
        help="pivot the view",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    if not args.jar:
        print("must specify jdbc driver jar(-J).")
        sys.exit(-1)

    if not args.driver:
        print("must specify jdbc driver name(-D).")
        sys.exit(-1)

    if not args.url:
        print("must specify connection string(-U).")
        sys.exit(-1)

    if not args.sql:
        print("must specify SQL statement(-Q).")
        sys.exit(-1)

    conn = jaydebeapi.connect(args.driver, args.url, jars=args.jar)
    cur = conn.cursor()
    cur.execute(args.sql)
    rchg = cur.rowcount
    if rchg == -1:
        try:
            data = cur.fetchall()
            hdr = [d[0] for d in cur.description]
            if args.debug:
                print("# header :", hdr)
                print("# data :", data)
            if args.pivot:
                print(SimpleTable(data=data, header=hdr).repr_pivot(), end="")
            else:
                if args.format == "json":
                    print(SimpleTable(data=data, header=hdr).get_json(), end="")
                elif args.format == "yaml":
                    print(SimpleTable(data=data, header=hdr).get_yaml(), end="")
                elif args.format == "csv":
                    print(SimpleTable(data=data, header=hdr).get_csv(), end="")
                else:
                    print(
                        SimpleTable(data=data, header=hdr, maxwidth=args.maxwidth),
                        end="",
                    )
        except:
            pass
        finally:
            cur.close()
    else:
        print("# {} rows impacted.".format(rchg))
    conn.close()


# postgresql
# sretools-jdbc-call  -J "/usr/share/java/postgresql.jar" -D "org.postgresql.Driver" -U "jdbc:postgresql://localhost:5432/sample?user=postgres&password=postgres" -Q "select id, count(*)cnt from t1 group by id"
# id cnt
# ------
# 1  1
# 2  1

# /jdbc_call.py -J ~/jdbc/db2jcc4.jar -D com.ibm.db2.jcc.DB2Driver  -U "jdbc:db2://localhost:50000/sample:user=db2inst1;password=db2inst1;" -Q "select viewschema,viewname,varchar(text) as text from syscat.views fetch first 1 rows only with ur"
# VIEWSCHEMA VIEWNAME          TEXT
# -------------------------------------------------------------------------------------------------------
# SYSIBM     CHECK_CONSTRAINTS CREATE OR REPLACE VIEW SYSIBM.CHECK_CONSTRAINTS
#                              (CONSTRAINT_CATALOG, CONSTRAINT_SCHEMA, CONSTRAINT_NAME, CHECK_CLAUSE)
#                              AS SELECT
#                              CAST(CURRENT SERVER AS VARCHAR(128)), TBCREATOR,
#                              CAST(NAME AS VARCHAR(128)), TEXT
#                              FROM SYSIBM.SYSCHECKS
#                              WHERE TYPE='C'
#                              UNION ALL
#                              SELECT CAST(CURRENT SERVER AS VARCHAR(128)), TBCREATOR,
#                              CAST(CONCAT(RTRIM(CONCAT(CHAR(CTIME), CHAR(FID) ) ),
#                              RTRIM(CHAR(COLNO)) ) AS VARCHAR(128) ),
#                              CAST(CONCAT(CONCAT('CHECK (', C.NAME), ' IS NOT NULL)') AS CLOB(64K) )
#                              FROM SYSIBM.SYSCOLUMNS C, SYSIBM.SYSTABLES T
#                              WHERE C.TBCREATOR = T.CREATOR AND C.TBNAME = T.NAME AND TYPE IN('U', 'T')
#                              AND NULLS ='N'
#

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang
# yhang_wang@icloud.com

import os
import sys
from optparse import OptionParser
import re
import getpass
import pexpect
import string
import datetime
import threading
import socket
from queue import Queue
from multiprocessing.pool import Pool
from random import randrange


def main():
    usage = "usage : python3 -m yssh [options] "
    parser = OptionParser(usage)
    parser.add_option(
        "-r",
        "--remote",
        dest="remote",
        help="remote servers. format of host1,host2,host3. Accepts host:port or user@host:port as well.",
    )
    parser.add_option("-c", "--command", dest="command", help="command to run remotely")
    parser.add_option("-f", "--script", dest="script", help="script to run remotely")
    parser.add_option("-u", "--user", dest="user", help="user")
    parser.add_option("-p", "--password", dest="password", help="password")
    parser.add_option(
        "-W", "--worker", dest="worker", default=4, help="max number of workers"
    )
    parser.add_option(
        "-C",
        "--call",
        dest="call",
        default="{{THEONE}}",
        help='way to call script, work with -f. eg. "python3 {{THEONE}}"',
    )
    parser.add_option("-i", "--pkey", dest="pkey", help="private key file for ssh")
    parser.add_option(
        "-w",
        "--workdir",
        dest="workdir",
        default=".cache/yssh",
        help="cache dir for script on remote server",
    )
    parser.add_option(
        "-t",
        "--timeout",
        dest="timeout",
        default=10,
        help="default timeout. better > 5",
    )
    parser.add_option(
        "-O", "--option", dest="option", default="", help="extra option for ssh command"
    )
    parser.add_option(
        "-I",
        "--init",
        dest="init",
        default="\r\n",
        help="command to run after ssh connection established.",
    )
    parser.add_option(
        "-X",
        "--debug",
        action="store_true",
        dest="debug",
        default=False,
        help="debug mode",
    )
    parser.add_option(
        "-R",
        "--remove",
        action="store_true",
        dest="remove",
        default=False,
        help="remove script and log after run",
    )
    parser.add_option(
        "-A",
        "--removeall",
        action="store_true",
        dest="removeall",
        default=False,
        help="with -R, remove all in remote cache",
    )
    parser.add_option(
        "-k",
        "--rm_conflict_key",
        action="store_true",
        dest="rm_conflict_key",
        default=False,
        help="remove known_hosts item without confirmation. this is useful when a large number of target server being rebuilt frequently.",
    )
    parser.add_option(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="show more info eg. command/script name",
    )
    parser.add_option(
        "-T",
        "--tmpfile",
        action="store_true",
        dest="tmpfile",
        default=False,
        help="use random temp name to avoid conflict",
    )

    (options, args) = parser.parse_args()

    options.timeout = int(options.timeout)
    if options.timeout <= 0:
        options.timeout = 10

    def sshrun(rserver, command):
        output = ""
        good = True
        if options.debug:
            print("# [{}] command =  {}".format(rserver, command))
        port = "22"
        sport = ""
        if ":" in rserver:
            rserver, port = re.split(":", rserver)[:2]
            if not port:
                port = "22"
            sport = "-p " + str(port)
        nullinput = " -n "
        if not (options.command or options.script):
            nullinput = " "
            cols, rows = os.get_terminal_size()
        sshbase = (
            "ssh "
            + "-o ConnectTimeout="
            + str(options.timeout)
            + nullinput
            + " "
            + options.option
            + " "
            + sport
        )
        if options.debug:
            print("# [{}] sshbase = {}".format(rserver, sshbase))
        if options.pkey:
            sshbase = sshbase + " -i " + options.pkey
        if "@" in rserver:
            sshbase += " " + rserver
        else:
            sshbase += " " + options.user + "@" + rserver
        if command:
            if options.debug:
                print("# [{}] {}".format(rserver, sshbase + ' "' + command + '"'))
            child = pexpect.spawn(
                sshbase + ' "' + command + '"', timeout=options.timeout
            )
        else:
            if options.debug:
                print("# [{}] {}".format(rserver, sshbase))
            child = pexpect.spawn(sshbase, timeout=options.timeout)
        printable = set(string.printable)
        old_r = -1
        good = True
        resent = False
        while True:
            r = child.expect(
                [
                    "want to continue connecting",
                    "assword:",
                    "ermission denied",
                    "[\$#]",
                    "key verification failed",
                    pexpect.TIMEOUT,
                    pexpect.EOF,
                ],
                timeout=options.timeout,
            )
            if r == 0:
                if options.debug:
                    print("# [{}] answer yes for continue connecting".format(rserver))
                child.sendline("yes")
            if r == 1:
                if options.debug:
                    print("# [{}] Requesting password".format(rserver))
                if old_r == 1:
                    print("# [{}] Repeated Requesting password".format(rserver))
                    child.close()
                    good = False
                    break
                child.sendline(options.password)
            if r == 2:
                if options.debug:
                    print("# [{}] wrong password".format(rserver))
                output += "# [{}] wrong password\n".format(rserver)
                child.close()
                good = False
                break
            if r == 3:
                if options.debug:
                    print("# [{}] $/#".format(rserver))
                child.sendline("\r\n")
                child.setwinsize(rows, cols)
                child.sendline(options.init)
                break
            if r == 4:
                content = child.before.decode()
                cmd = [
                    ln.strip()
                    for ln in content.splitlines()
                    if re.match(r"\s*ssh-keygen -f.*?-R.*", ln)
                ]
                if options.debug:
                    print(
                        "# [{}] key validation failed. target server may have been rebuilt.".format(
                            rserver
                        )
                    )
                    print(
                        "# [{}] pls check and remove needed from known_hosts.".format(
                            rserver
                        )
                    )
                else:
                    output += "# [{}] key validation failed. target server may have been rebuilt.\n".format(
                        rserver
                    )
                    output += "# [{}] pls check and remove needed from known_hosts.\n".format(
                        rserver
                    )
                if options.rm_conflict_key:
                    if options.debug:
                        print("# [{}] fixing known_hosts".format(rserver))
                    else:
                        output += "# [{}] fixing known_hosts\n".format(rserver)
                    if len(cmd) > 0:
                        for c in cmd:
                            os.system(c)
                    else:
                        os.system(
                            "ssh-keygen -f ~/.ssh/known_hosts -R {} > /dev/null 2>&1 ".format(
                                rserver
                            )
                        )
                else:
                    break
                if options.debug:
                    print("# resending request ...")
                child.close()
                if command:
                    if options.debug:
                        print(
                            "# [{}] {}".format(rserver, sshbase + ' "' + command + '"')
                        )
                    child = pexpect.spawn(
                        sshbase + ' "' + command + '"', timeout=options.timeout
                    )
                else:
                    if options.debug:
                        print("# [{}] {}".format(rserver, sshbase))
                    child = pexpect.spawn(sshbase, timeout=options.timeout)
                resent = True
            if r == 5:
                output += "# [{}] TIMEOUT during [{}]\n".format(rserver, command)
                child.close()
                good = False
                break
            if r == 6:
                if options.debug:
                    print("# [{}] EOF detected".format(rserver))
                if old_r == 4 and resent:
                    resent = False
                content = child.before.decode()
                if re.search(r"\S+", content):
                    lines = content.splitlines()
                    if re.search(r"\S+", lines[0]):
                        start = 0
                    else:
                        start = 1
                    if re.search(r"\S+", lines[-1]):
                        lines = lines[start:]
                    else:
                        lines = lines[start:-1]
                    for ln in lines:
                        output += ("".join(filter(lambda x: x in printable, ln))) + "\n"
                break
            old_r = r
        child.interact()
        return (output, good)

    def scprun(command):
        if options.debug:
            print("# [scp] {}".format(command))
        output = ""
        child = pexpect.spawn(command, timeout=options.timeout)
        while True:
            """
            no key validation error here. as even when used as scp alternative,
            sshrun/mkdir is called first.
            """
            r = child.expect(
                [
                    "want to continue connecting",
                    "assword:",
                    "ermission denied",
                    "[\$#]",
                    pexpect.TIMEOUT,
                    pexpect.EOF,
                ],
                timeout=options.timeout,
            )
            if r == 0:
                if options.debug:
                    print("# [scp] answer yes")
                child.sendline("yes")
            if r == 1:
                if options.debug:
                    print("# [scp] paste password")
                child.sendline(options.password)
            if r == 2:
                output += "# [scp] Wrong password or permission. cannot login.\n"
                return False
            if r == 3:
                return True
            if r == 4:
                output += "# [scp] Timeout during [{}]\n".format(command)
                return False
            if r == 5:
                content = child.before.decode()
                if re.search(r"\S+", content):
                    print(content)
                break
        return True

    def sshfile(rserver, script):
        rsvr, port = re.split(":", rserver)
        luser = options.user
        if "@" in rsvr:
            luser, rsvr = re.split(r"@", rsvr)[:2]
        output = ""
        if not re.search(r"^/", options.workdir):
            o, good = sshrun(
                rserver, "mkdir -p ~" + options.user + "/" + options.workdir
            )
            if not good:
                output += "# [{}] sshfile exited.\n".format(rserver)
                return output
        else:
            o, good = sshrun(rserver, "mkdir -p " + options.workdir)
            if not good:
                output += "# [{}] sshfile exited.\n".format(rserver)
                return output
        scpbase = "scp -q -P {} -o ConnectTimeout={} ".format(port, options.timeout)
        if options.pkey:
            scpbase = scpbase + " -i " + options.pkey
        basename = re.sub(r".*\/", "", script)

        def randstr():
            allchar = string.ascii_letters + string.digits
            result = "".join(allchar[randrange(len(allchar))] for _ in range(30))
            return result

        if options.tmpfile:
            tmstr = "{date:%Y%m%d.%H%M%S}".format(date=datetime.datetime.now())
            randname = "_" + tmstr + "_" + randstr()
        else:
            randname = ""
        if re.search(r"^/", options.workdir):
            tgtfile = options.workdir + "/" + basename + randname
        else:
            tgtfile = "~" + luser + "/" + options.workdir + "/" + basename + randname
        tgtlog = tgtfile + ".log"
        go2workdir = ""
        if os.path.isfile(script):
            if not scprun(
                scpbase + " " + script + " " + luser + "@" + rsvr + ":" + tgtfile
            ):
                return output
        elif os.path.isdir(script):
            sshrun(rserver, "mkdir -p {}".format(tgtfile))
            go2workdir = "cd {}/{};".format(tgtfile, basename)
            if not scprun(
                scpbase + " -r " + script + " " + luser + "@" + rsvr + ":" + tgtfile
            ):
                return output
            # sshrun(rserver, 'chmod -R 0755 ' + tgtfile)
        if os.path.isfile(script):
            tgtcmd = re.sub(r"\{\{THEONE\}\}", tgtfile, options.call)
        elif os.path.isdir(script):
            tgtcmd = re.sub(r"\{\{THEONE\}\}", tgtfile + "/" + basename, options.call)
        if options.debug:
            print("# [{}] cmd = {}".format(rserver, tgtcmd))
        mnohup = re.search(r"^\s*nohup", tgtcmd)
        mredirect = re.search(r"\>", tgtcmd) or re.search(r"\&", tgtcmd)
        if mnohup or mredirect:
            res, _ = sshrun(
                rserver, go2workdir + "chmod -R 0755 " + tgtfile + ";  " + tgtcmd
            )
        else:
            res, _ = sshrun(
                rserver,
                go2workdir
                + "chmod -R 0755 "
                + tgtfile
                + ";  "
                + tgtcmd
                + " 2>&1 | tee "
                + tgtlog,
            )
        if res:
            output += res
        if options.remove:
            if options.removeall:
                res, _ = sshrun(
                    rserver, "rm -rf " + "~" + luser + "/" + options.workdir + "/" + "*"
                )
            else:
                res, _ = sshrun(rserver, "rm -rf " + tgtfile + "*")
            if res:
                output += res
        return output

    if not options.remote:
        parser.print_help()
        sys.exit(-1)

    if not options.user:
        options.user = getpass.getuser()
    if not options.pkey and not options.password:
        options.password = getpass.getpass()

    if options.script and not (
        os.path.isfile(options.script) or os.path.isdir(options.script)
    ):
        print("# {} does not exist".format(options.script))
        sys.exit(-1)

    TASK_QUEUE = Queue()
    RESULT_QUEUE = Queue()
    END_OF_ALL = "__yx_end_of_work__"

    servers = [x for x in re.split(r",", options.remote) if x]
    for rserver in servers:
        TASK_QUEUE.put(rserver)
    if not options.script and not options.command:
        NUM_WORKER_THREAD = 1
    else:
        NUM_WORKER_THREAD = int(options.worker)
        if NUM_WORKER_THREAD > len(servers):
            NUM_WORKER_THREAD = len(servers)
    if options.debug:
        print("# using {} worker thread(s).".format(NUM_WORKER_THREAD))

    for _ in range(NUM_WORKER_THREAD):
        TASK_QUEUE.put(END_OF_ALL)

    def worker(id=-1):
        def porttest(host, port):
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.settimeout(options.timeout)
                s.connect((host, port))
            except:
                return False
            return True

        def runsvr(rsvr):
            output = "# HOST : [" + rsvr + "]\n"
            port = "22"
            if ":" in rsvr:
                rsvr, port = re.split(":", rsvr)[:2]
                if not port:
                    port = "22"
            luser = None
            if "@" in rsvr:
                luser, rsvr = re.split("@", rsvr)[:2]
            if options.debug:
                print(
                    "# [worker {}] rsvr={}, luser={}, port={}".format(
                        id, rsvr, luser, port
                    )
                )
            if not porttest(rsvr, int(port)):
                output += "# [worker {}] [{}] ssh port {} not reachable\n".format(
                    id, rsvr, port
                )
                return output
            if luser:
                rsvr = luser + "@" + rsvr + ":" + port
            else:
                rsvr = rsvr + ":" + port
            if options.debug:
                print("# [worker {}] started working for {}".format(id, rsvr))
            if not re.search(r"\S+", rsvr):
                return
            if options.command:
                if options.verbose:
                    output += "# [worker {}] command = {}\n".format(id, options.command)
                res, _ = sshrun(rsvr, options.command)
                output += res
            if options.script:
                if options.verbose:
                    output += "# [worker {}] script  = {}\n".format(id, options.script)
                output += sshfile(rsvr, options.script)
            if not options.script and not options.command:
                res, _ = sshrun(rsvr, None)
                output += res
            return output

        while True:
            rsvr = TASK_QUEUE.get()
            if rsvr == END_OF_ALL:
                break
            res = runsvr(rsvr)
            RESULT_QUEUE.put(res)
        RESULT_QUEUE.put(END_OF_ALL)
        if options.debug:
            print("# worker {} exited.".format(id))

    def collector():
        ended = 0
        while True:
            res = RESULT_QUEUE.get()
            if res == END_OF_ALL:
                ended += 1
                if ended >= NUM_WORKER_THREAD:
                    break
            else:
                print(res)

    threads = list()
    c = threading.Thread(target=collector)
    c.start()
    threads.append(c)
    for i in range(NUM_WORKER_THREAD):
        t = threading.Thread(target=worker, args=(i,))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import re
import csv
import copy
from wcwidth import wcswidth
from collections import defaultdict
from itertools import zip_longest
from sretools import SimpleTable


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n",
        "--lineno",
        dest="lineno",
        default=False,
        action="store_true",
        help="show line no.",
    )
    parser.add_argument("-H", "--header", dest="header", help="header columns")
    parser.add_argument("-f", "--infile", dest="infile", help="input file")
    parser.add_argument(
        "-C",
        "--dump-column",
        dest="dumpcols",
        help="only print columns indentified by index numbers.",
    )
    parser.add_argument(
        "-s", "--sortby", dest="sortby", help="column id starts with 0."
    )
    parser.add_argument(
        "-d",
        "--delimiter",
        dest="delimiter",
        default=",",
        help="char to seperate columns",
    )
    parser.add_argument(
        "-q",
        "--quotechar",
        dest="quotechar",
        default='"',
        help="char to seperate columns",
    )
    parser.add_argument(
        "-w",
        "--maxwidth",
        dest="maxwidth",
        type=int,
        default=-1,
        help="max col width when print in console, min==20",
    )
    parser.add_argument(
        "-v",
        "--pivot",
        dest="pivot",
        action="store_true",
        default=False,
        help="pivot wide tables.",
    )
    parser.add_argument(
        "-F", "--format", dest="format", help="other than console table, json,yaml,html"
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    header = list()
    data = list()
    if args.lineno:
        header.append("#")
    headered = False
    if args.header:
        header += re.split(args.delimiter, args.header)
        headered = True
    if not args.infile:
        INPUT = sys.stdin
    else:
        INPUT = open(args.infile, "r", newline="")
    # with open(args.infile,"r",newline='') as f :
    if True:
        cr = csv.reader(INPUT, delimiter=args.delimiter, quotechar=args.quotechar)
        if not headered:
            header += next(cr)
            headered = True
        for r in cr:
            data.append(r)
    INPUT.close()

    if args.sortby:

        def fsort(x):
            v = list()
            for i in re.split(r",", args.sortby):
                v0 = x[int(i)] or ""
                if re.match(r"^\d+(\.\d*)*$", v0):
                    v.append(float(v0))
                else:
                    v.append(0)
                v.append(v0)
            return v

        data = sorted(data, key=fsort)
    if args.pivot:
        print(SimpleTable(header=header, data=data, cols=args.dumpcols).repr_pivot())
    else:
        t = SimpleTable(
            header=header, data=data, cols=args.dumpcols, maxwidth=args.maxwidth
        )
        if args.format == "json":
            print(t.get_json(), end="")
        elif args.format == "yaml":
            print(t.get_yaml(), end="")
        elif args.format == "html":
            print(t.get_html(), end="")
        else:
            print(t, end="")


if __name__ == "__main__":
    main()



from .ltexpect import LE
from .tblfmt import SimpleTable
from .json2html import JsonConverter
#from .dbx import DBX
from .commandline import commandline

#__all__ = [LE, SimpleTable, JsonConverter, DBX, commandline]
__all__ = [LE, SimpleTable, JsonConverter, commandline]



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import json
import xmltodict


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--infile", dest="infile", help="input file")
    parser.add_argument(
        "-s",
        "--sorted",
        dest="sorted",
        action="store_true",
        default=False,
        help="sort keys",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    if args.infile:
        if not os.path.isfile(args.infile):
            print("{} not exists.")
        with open(args.infile, "r") as f:
            INPUT = f.read()
    else:
        INPUT = sys.stdin.read()

    try:
        yl = xmltodict.parse(INPUT)
    except:
        print("invalid or unrecognized XML")
        if args.debug:
            print(INPUT)
        sys.exit(-1)

    print(json.dumps(yl, sort_keys=args.sorted, indent=2))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import re
import jaydebeapi
import socket
import psutil
import traceback
import json
import time
import getpass
import yaml

# from multiprocessing import Process as Task
from threading import Thread as Task
from sretools import SimpleTable


class DBX:
    def __init__(self, jar, driver, url, user, password):
        self.__jar = jar
        self.__driver = driver
        self.__url = url
        self.__user = user
        self.__password = password
        self.__conn = None
        self.__cursor = None

    def close(self):
        try:
            self.__cursor.close()
            self.__conn.close()
        except:
            pass

    def get_connection(self):
        if not (self.__conn and self.__cursor):
            if (
                not self.__password
                or re.match(r"ignore", self.__password, re.IGNORECASE)
                or re.search(r"password=", self.__url)
            ):
                auth = None
            else:
                auth = [self.__user, self.__password.encode()]
            self.__conn = jaydebeapi.connect(
                self.__driver, self.__url, auth, jars=self.__jar
            )
            self.__cursor = self.__conn.cursor()

    def run_sql(self, sql):
        result = dict()
        try:
            self.__cursor.execute(sql)
            rchg = self.__cursor.rowcount
            result["sql"] = sql
            result["rows_impacted"] = rchg
            if rchg == -1:
                data = list()
                for row in self.__cursor.fetchall():
                    data.append([str(c) if c is not None else "" for c in row])
                hdr = [d[0] for d in self.__cursor.description]
                result["header"] = hdr
                result["data"] = data
        # except (Exception, jaydebeapi.DatabaseError) :
        except:
            result = {"error": traceback.format_exc().splitlines()[-1], "sql": sql}
        return json.dumps(result)

    def run_sp(self, sp, args=list()):
        pass


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-J", "--jar", dest="jar", help="jdbc driver jar file")
    parser.add_argument(
        "-P", "--plugin", dest="plugin", help="plugin file for user defined commands"
    )
    parser.add_argument(
        "-u", "--user", dest="user", help="user. default as current unix user."
    )
    parser.add_argument(
        "-p", "--password", dest="password", help='password. ENV["dbx_mypwd"]'
    )
    parser.add_argument("-D", "--driver", dest="driver", help="jdbc driver name")
    parser.add_argument("-U", "--url", dest="url", help="jdbc url or connection string")
    parser.add_argument(
        "-C",
        "--connect",
        dest="connect",
        action="store_true",
        default=False,
        help="force connect as server",
    )
    parser.add_argument("-Q", "--sql", dest="sql", help="sql statement to run")
    parser.add_argument(
        "-w",
        "--maxwidth",
        dest="maxwidth",
        type=int,
        default=-1,
        help="maxwidth of column",
    )
    parser.add_argument(
        "-F", "--outformat", dest="format", default="", help="json,yaml,csv,html"
    )
    parser.add_argument(
        "-v",
        "--pivot",
        dest="pivot",
        action="store_true",
        default=False,
        help="pivot the view",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    ppid = os.getppid()
    found = False
    p = ppid
    for d in ["~/.cache", "~/.cache/sretools"]:
        fulld = os.path.expanduser(d)
        if not os.path.exists(fulld):
            os.mkdir(fulld)
    while p not in [0, 1]:
        if os.path.exists(os.path.expanduser("~/.cache/sretools/.dbx.{}".format(p))):
            found = True
            break
        else:
            np = psutil.Process(ppid).ppid()
            if p == np:
                break
            else:
                p = np

    global CMDDT
    if args.plugin and os.path.isfile(args.plugin):
        with open(args.plugin, "r") as f:
            CMDDT = yaml.safe_load(f.read())
    else:
        CMDDT = dict()

    if not found or args.connect:
        addr = os.path.expanduser("~/.cache/sretools/.dbx.{}".format(ppid))
        if os.path.exists(addr):
            if args.debug:
                print("# ready to remove old server process")
            fdummy = addr + ".pid"
            for p in psutil.process_iter():
                try:
                    if fdummy in str(p.open_files()):
                        p.kill()
                except:
                    pass
            os.unlink(addr)
            if os.path.exists(fdummy):
                os.unlink(fdummy)
    else:
        # client only
        addr = os.path.expanduser("~/.cache/sretools/.dbx.{}".format(p))

    def get_sql_command(sqlorpath):
        global CMDDT
        if args.debug:
            print("# sqlorpath = ", sqlorpath)
            print("# CMDDT = ", CMDDT)
        cmddt = CMDDT
        if not cmddt or not sqlorpath:
            return sqlorpath
        if not re.search(r"^\s*\\", sqlorpath):
            return sqlorpath
        s = re.sub(r"^\s*\\", "", sqlorpath)
        p = cmddt
        for c in re.split(r"\.", s.strip()):
            if c in p:
                p = p[c]
            else:
                return sqlorpath
        return p

    def clientrun():
        rsql = get_sql_command(args.sql)
        try:
            if args.debug:
                print("# DBXClient {}".format(addr))
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            if args.connect and rsql:
                attempt = 10
                while attempt > 0:
                    if args.debug:
                        print("attempt = {}".format(attempt))
                    try:
                        time.sleep(1)
                        s.connect(addr)
                        break
                    except:
                        attempt -= 1
                if attempt == 0:
                    print("# cannot establish connections.")
                    return -1
            else:
                try:
                    s.connect(addr)
                except:
                    print("# cannot establish connections.")
                    return -1
            s.send(rsql.encode())
            if args.debug:
                print("# Client sent : {}".format(rsql))
            rsp = s.recv(1024 * 1024).decode()
            if args.debug:
                print("# Client got :", rsp)
            try:
                obj = json.loads(rsp)
                if "error" in obj:
                    print(obj.get("error", ""))
                    return -1
            except:
                obj = dict()
            if (
                not re.match(r"^\s*\\q\s*$", args.sql)
                and "header" in obj
                and "data" in obj
            ):
                if args.format == "json":
                    t = SimpleTable(header=obj["header"], data=obj["data"])
                    print(t.get_json(), end="")
                elif args.format == "yaml":
                    t = SimpleTable(header=obj["header"], data=obj["data"])
                    print(t.get_yaml(), end="")
                elif args.format == "csv":
                    t = SimpleTable(header=obj["header"], data=obj["data"])
                    print(t.get_csv(), end="")
                elif args.format == "html":
                    t = SimpleTable(header=obj["header"], data=obj["data"])
                    print(t.get_html(), end="")
                else:
                    t = SimpleTable(
                        header=obj["header"], data=obj["data"], maxwidth=args.maxwidth
                    )
                    if args.pivot:
                        print(t.repr_pivot(), end="")
                    else:
                        print(t, end="")
            rw = obj.get("rows_impacted", 0)
            if rw >= 0 and not re.match(r"^\s*\\q\s*$", args.sql):
                print("# {} row(s) impacted.".format(rw))
        except:
            print(traceback.format_exc().splitlines()[-1])

    # server
    if args.connect:
        if not args.user:
            args.user = getpass.getuser()
        if not args.password:
            args.password = os.environ.get("dbx_password", None)
            if not args.password and not re.search(r"password=", args.url):
                args.password = getpass.getpass(
                    prompt="# password for {} :".format("args.user")
                )
        if (
            not args.password
            or (not re.search(r"\S+", args.password))
            or re.match(r"ignore", args.password, re.IGNORECASE)
        ):
            args.password = ""
        if not all([args.jar, args.driver, args.url]):
            print("# Must specify jdbc driver jar(-J), driver(-D) and URL(-U).")
            return -1
        pid = os.fork()
        # keep attachment to pid file
        fdummy = open(addr + ".pid", "w")
        # child
        if pid == 0:
            print("# DBXServer {}@{}".format(os.getpid(), addr))
            args.driver = os.path.expanduser(args.driver)
            dbx = DBX(
                driver=args.driver,
                url=args.url,
                jar=args.jar,
                user=args.user,
                password=args.password,
            )
            dbx.get_connection()
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.settimeout(600)
            s.bind(addr)
            s.listen(1)

            def sqlworker(netconn, osql):
                try:
                    sql = get_sql_command(osql)
                    result = dbx.run_sql(sql)
                    if args.debug:
                        print("# Sqlworker: sql={}, result={}".format(sql, result))
                    netconn.send(result.encode())
                except:
                    print(traceback.format_exc().splitlines()[-1])

            while True:
                try:
                    netconn, cltaddr = s.accept()
                    sql = netconn.recv(32768).decode()
                    if args.debug:
                        print("# Eventloop: sql = [{}]".format(sql))
                    if re.match(r"^\s*\\q\s*$", sql):
                        break
                    wk = Task(target=sqlworker, args=(netconn, sql))
                    wk.start()
                except:
                    msg = traceback.format_exc()
                    print(msg.splitlines()[-1:])
                    if re.search("timeout", msg, re.IGNORECASE):
                        break
            if os.path.exists(addr):
                os.unlink(addr)
        # parent
        elif pid > 0:
            if args.sql:
                clientrun()
            return 0
        else:
            print("# Error forking new process ...")
            return -1
    # client
    else:
        if not args.sql:
            s = ""
            for ln in sys.stdin:
                if not ln or re.search(r"^\s*#", ln) or re.search(r"^\-\-", ln):
                    continue
                s += " " + ln.rstrip()
            args.sql = s.strip()
        clientrun()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import json
import yaml
from sretools import JsonConverter


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--infile", dest="infile", help="input file")
    parser.add_argument(
        "-s",
        "--sorted",
        dest="sorted",
        action="store_true",
        default=False,
        help="sort keys",
    )
    parser.add_argument(
        "-a",
        "--table_attributes",
        dest="attributes",
        default="border-collapse:collapse;",
        help="table attributes",
    )
    parser.add_argument(
        "-C", "--colors", dest="colors", help="colors to be used for background"
    )
    parser.add_argument(
        "-I",
        "--keys-included",
        dest="keys_included",
        action="append",
        help="include these keys only",
    )
    parser.add_argument(
        "-E",
        "--keys-excluded",
        dest="keys_excluded",
        action="append",
        help="exclude these keys",
    )
    parser.add_argument(
        "-k",
        "--keywords_order",
        dest="keyorder",
        help="use this ordered keywords to help sort dict.",
    )
    parser.add_argument(
        "-v",
        "--dictsortbyvalue",
        dest="dtbyval",
        action="store_true",
        default=False,
        help="with -s, sort dict by value.",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    if args.infile:
        if not os.path.isfile(args.infile):
            print("{} not exists.")
        with open(args.infile, "r") as f:
            INPUT = f.read()
    else:
        INPUT = sys.stdin.read()

    try:
        yl = yaml.safe_load(INPUT)
    except:
        print("not invalid YAML")
        if args.debug:
            print(INPUT)
        sys.exit(-1)

    print(
        JsonConverter(
            jsstr=yl,
            tblattr=args.attributes,
            sort=args.sorted,
            sort_by_val=args.dtbyval,
            sortkeywords=args.keyorder,
            keys_included=args.keys_included,
            keys_excluded=args.keys_excluded,
            colors=args.colors,
        ).json2html()
    )


if __name__ == "__main__":
    main()



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="sretools",
    version="0.1.57",
    description="SRE console/terminal toolbox",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/laowangv5/sretools",
    author="Yonghang Wang",
    author_email="wyhang@gmail.com",
    license="MIT",
    classifiers=["License :: OSI Approved :: MIT License"],
    packages=["sretools"],
    include_package_data=True,
    install_requires=[
        "pexpect",
        "texttable",
        "wcwidth",
        "pyyaml",
        "jaydebeapi",
        "psutil",
        "xmltodict",
        "dicttoxml",
    ],
    keywords=[
        "sretool",
        "sre",
        "ssh",
        "yssh",
        "pssh",
        "ltexpect",
        "expect",
        "incident",
        "postmortem",
        "json2html",
        "xml2json",
        "json2xml",
        "json",
        "jsonfmt",
        "html",
        "yaml",
        "jdbc",
        "dbx",
        "dbquery",
        "qx",
    ],
    entry_points={
        "console_scripts": [
            "sretools-ssh=sretools.yssh:main",
            "sretools-expect=sretools.ltexpect:main",
            "sretools-nonascii=sretools.nonascii:main",
            "sretools-table-format=sretools.tblfmt:main",
            "sretools-csv-format=sretools.csvfmt:main",
            "sretools-json2html=sretools.json2html:main",
            "sretools-jsonfmt=sretools.jsonfmt:main",
            "sretools-json2yaml=sretools.json2yaml:main",
            "sretools-yaml2json=sretools.yaml2json:main",
            "sretools-xml2json=sretools.xml2json:main",
            "sretools-json2xml=sretools.json2xml:main",
            "sretools-yaml2html=sretools.yaml2html:main",
            "sretools-dbx=sretools.dbx:main",
        ]
    },
)



#!/usr/bin/env python3
# Yonghang Wang

import sys
import argparse
import os
import re
import csv
import copy
from wcwidth import wcswidth
from collections import defaultdict
from itertools import zip_longest


class SimpleTable:
    def __init__(self, data=None, header=None, cols=None, maxwidth=-1, noheader=False):
        self.__maxwidth = int(maxwidth)
        self.__noheader = noheader
        if cols and re.search(",", cols):
            ncols = len(header)
            xmap = [int(i) for i in re.split(r",", cols) if int(i) < ncols]
            self.__header = [header[i] for i in xmap]
            self.__data = list()
            for d in data:
                self.__data.append([d[i] for i in xmap])
        else:
            self.__header = header or list()
            self.__data = data or list()
            if self.__noheader and len(self.__data) > 0 :
                self.__header = self.__data[0]
        self.__num_of_cols = len(self.__header)

    def import_csv(self, csvfile, header=True):
        pass

    def __len__(self):
        return len(self.__data)

    def __getitem__(self, pos):
        return self.__data[pos]

    def __add__(self, other):
        if self.__num_of_cols != other.__num_of_cols:
            return None
        return SimpleTable(self.__data + other.__data, self.__header)

    def __iadd__(self, other):
        if self.__num_of_cols != other.__num_of_cols:
            return None
        self.__data += other.__data

    def get_json(self):
        import json

        tbl = list()
        for r in self.__data:
            row = dict()
            for ix, val in enumerate(r):
                row[self.__header[ix]] = val
            tbl.append(row)
        return json.dumps(tbl, indent=2)

    def get_yaml(self):
        import yaml

        tbl = list()
        for r in self.__data:
            row = dict()
            for ix, val in enumerate(r):
                row[self.__header[ix]] = val
            tbl.append(row)
        return yaml.safe_dump(tbl, default_flow_style=False)

    def get_csv(self):
        import io

        si = io.StringIO()
        w = csv.writer(si)
        if self.__header:
            w.writerow(self.__header)
        for r in self.__data:
            w.writerow(r)
        return si.getvalue()

    def get_html(self):
        res = '<table border=1 style="border-collapse:collapse;">\n'
        res += "<tr>\n"
        for h in self.__header:
            res += "<td><b>" + str(h) + "</b></td>\n"
        res += "</tr>\n"
        for row in self.__data:
            res += "<tr>\n"
            for c in row:
                if not c :
                    if c == 0 :
                        c = "0"
                    else :
                        c = ""
                res += "<td>" + str(c.replace("\n", "<br>")) + "</td>\n"
            res += "</tr>\n"
        return res

    def repr_pivot(self):
        maxwidth = max([wcswidth(h) for h in self.__header])
        fmtstr = "{:" + str(maxwidth) + "} : {}"
        res = ""
        for r, row in enumerate(self.__data):
            if r > 0:
                res += "--\n"
            for i, c in enumerate(row):
                res += fmtstr.format(self.__header[i], c) + "\n"
        return res

    def __wcswidth_x(self, s):
        res = 0
        for ln in s.splitlines():
            wclen = wcswidth(ln)
            if wclen > res:
                res = wclen
        return res

    def __splitstring(self, s, maxwidth):
        result = list()
        for ln in s.splitlines():
            if wcswidth(ln) <= maxwidth:
                result.append(ln)
            else:
                left = ln
                while left:
                    start = 0
                    end = min(maxwidth, len(left))
                    while wcswidth(left[start:end]) > maxwidth:
                        end -= 1
                    result.append(left[start:end])
                    left = left[end:]
        return result

    def __splitrow(self, row):
        if self.__maxwidth < 20:
            return zip_longest(*[c.splitlines() for c in row], fillvalue="")
        else:
            return zip_longest(
                *[self.__splitstring(str(c), self.__maxwidth) for c in row],
                fillvalue=""
            )

    def __repr__(self):
        width = [0 for _ in range(len(self.__header))]
        for row in self.__data + [self.__header]:
            for ix, col in enumerate(row):
                if not col :
                    if col == 0 :
                        col = "0"
                    else :
                        col = ""
                if ix > len(width) - 1:
                    break
                wclen = self.__wcswidth_x(str(col))
                if wclen > width[ix]:
                    width[ix] = wclen
        if self.__maxwidth >= 20:
            width = [min(w, self.__maxwidth) for w in width]
        twidth = copy.copy(width)
        for ix, w in enumerate(twidth):
            twidth[ix] = (
                w - wcswidth(str(self.__header[ix])) + len(str(self.__header[ix]))
            )
        fmtstr = "".join(["{:" + str(l + 1) + "}" for l in twidth])
        res = ""
        if not self.__noheader:
            res += fmtstr.format(*self.__header).rstrip() + "\n"
            res += ("-" * (sum(width) + len(width) - 1)) + "\n"
        for r in self.__data:
            row = [str(c) if c or c==0 else "" for c in r]
            if len(row) < len(width):
                row.extend([""] * (len(width) - len(row)))
            for t in self.__splitrow(row):
                twidth = copy.copy(width)
                for ix, w in enumerate(twidth):
                    twidth[ix] = w - wcswidth(str(t[ix])) + len(str(t[ix]))
                fmtstr = "".join(["{:" + str(l + 1) + "}" for l in twidth])
                res += fmtstr.format(*t).rstrip() + "\n"
        return res


def tokenize(s):
    offset = 0
    tokens = list()
    while True:
        m = re.search(r"\S+\s*", s[offset:])
        if m:
            tokens.append([offset + m.start(), offset + m.end(), m.group().rstrip()])
            offset += m.end()
        else:
            break
    tokens[-1][1] = -1
    return tokens


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n",
        "--lineno",
        dest="lineno",
        default=False,
        action="store_true",
        help="show line no.",
    )
    parser.add_argument("-H", "--header", dest="header", help="header columns")
    parser.add_argument("-f", "--infile", dest="infile", help="input file")
    parser.add_argument(
        "-c",
        "--column",
        action="append",
        dest="column",
        help="column names. used when there" "re spaces within.",
    )
    parser.add_argument(
        "-C",
        "--dump-column",
        dest="dumpcols",
        help="only print columns indentified by index numbers.",
    )
    parser.add_argument(
        "-s", "--sortby", dest="sortby", help="column id starts with 0."
    )
    parser.add_argument(
        "-b",
        "--sep-char",
        dest="sepchar",
        default="\s+",
        help="char to seperate columns",
    )
    parser.add_argument(
        "-w",
        "--maxwidth",
        dest="maxwidth",
        type=int,
        default=-1,
        help="max col width when print in console, min==20",
    )
    parser.add_argument(
        "-t",
        "--table",
        dest="table",
        action="store_true",
        default=False,
        help="input preformatted by spaces. header should not include spaces.",
    )
    parser.add_argument(
        "-v",
        "--pivot",
        dest="pivot",
        action="store_true",
        default=False,
        help="pivot wide tables.",
    )
    parser.add_argument("-F", "--format", dest="format", help="json,yaml,csv,html")
    parser.add_argument(
        "-d",
        "--dataonly",
        dest="dataonly",
        action="store_true",
        default=False,
        help="indicate there's no header",
    )
    parser.add_argument(
        "-X",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="debug mode",
    )
    args = parser.parse_args()

    colsdict = defaultdict()
    colsdict_revert = defaultdict()
    if args.column:
        for c in args.column:
            oc = c
            mc = re.sub("\s", "_", oc)
            colsdict[oc] = mc
            colsdict_revert[mc] = oc
    if args.header and colsdict:
        for c, nc in colsdict.items():
            args.header = re.sub(c, nc, args.header)
    xheader = None
    header = list()
    if args.lineno:
        header.append("#")
    if args.header:
        mhdr = args.header
        if not args.table:
            header.extend(re.split(r"{}".format(args.sepchar), mhdr))
        else:
            xheader = tokenize(mhdr)
            header = [v.rstrip() for (s, t, v) in xheader]
    data = list()
    lno = 0
    INPUT = sys.stdin
    if args.infile:
        with open(args.infile, "r") as f:
            INPUT = f.readlines()
    for ln in INPUT:
        if not ln or not re.search(r"\S+", ln):
            continue
        if not header:
            for c, nc in colsdict.items():
                ln = re.sub(c, nc, ln)
        if not args.table:
            arr = re.split(r"{}".format(args.sepchar), ln.rstrip())
        else:
            if xheader:
                arr = [ln[s:t].rstrip() for (s, t, v) in xheader]
            else:
                for c, nc in colsdict.items():
                    ln = re.sub(c, nc, ln)
                xheader = tokenize(ln)
                if args.lineno:
                    header = ["#"] + [v.rstrip() for (s, t, v) in xheader]
                else:
                    header = [v.rstrip() for (s, t, v) in xheader]
                continue
        if len(header) == 0 or (args.lineno and len(header) == 1):
            if not args.table:
                header.extend(arr)
        else:
            lno += 1
            if args.lineno:
                data.append([str(lno)] + arr)
            else:
                data.append(arr)
    oheader = list()
    for h in header:
        nh = colsdict_revert.get(h, h)
        oheader.append(nh)
    if args.dataonly:
        data = [oheader] + data
    if args.sortby:

        def fsort(x):
            v = list()
            for i in re.split(r",", args.sortby):
                v0 = x[int(i)] or ""
                if re.match(r"^\d+(\.\d*)*$", v0):
                    v.append(float(v0))
                else:
                    v.append(0)
                v.append(v0)
            return v

        data = sorted(data, key=fsort)
    if args.pivot:
        print(SimpleTable(header=oheader, data=data, cols=args.dumpcols).repr_pivot())
    else:
        t = SimpleTable(
            header=oheader,
            data=data,
            cols=args.dumpcols,
            maxwidth=args.maxwidth,
            noheader=args.dataonly,
        )
        if args.format == "json":
            print(t.get_json(), end="")
        elif args.format == "yaml":
            print(t.get_yaml(), end="")
        elif args.format == "csv":
            print(t.get_csv(), end="")
        elif args.format == "html":
            print(t.get_html(), end="")
        else:
            print(t, end="")


if __name__ == "__main__":
    main()
